
////////////////////////////////////////////////////////
//// AUTHOR: YangYunLong1988
//// SOURCE: https://github.com/YangYunLong1988/w5500
////////////////////////////////////////////////////////


module SPI(
            input clk,		
            input rst_n,
            input miso,
            input [71:0] tx_data,           // The data to be sent by SPI is the largest frame of data, which is 6 bytes at most.
            input tx_en,                    // Send enable, high effective, as a flag to trigger the send state machine
            input rx_en,                    // Receive enable, high effective, as a flag to trigger the receive state machine
            input [2:0] tx_len,             // Send several bytes of data
            input [1:0] rx_len,             // Read several bytes
            input [15:0] rx_size,           // The number of bytes to read
             
            output sclk,                     // SPI Clock
            output reg mosi,
            output reg [7:0] rx_data_1,     // One byte of data received by SPI
            output reg [15:0] rx_data_2,    // Two bytes of data received by SPI
            output reg tx_ready,			// A frame of data transmission completion flag
            output reg rx_ready,			// 1/2 Data reception completion flag
            output reg sel,
            output reg rx_flag,
            output reg [7:0] buff_8
          );


    parameter CLKSPI = 9'd500;

    //Capture the rising edge of tx_en, reset the cnt count, and start generating the SPI clock
    reg tx_en_1;
    reg tx_en_2;
    reg tx_en_3;
    always @(posedge clk or negedge rst_n)
        if (!rst_n)
            begin
                tx_en_1 <= 1'b0;
                tx_en_2 <= 1'b0;
                tx_en_3 <= 1'b0;
            end
        else 
            begin 
                tx_en_1 <= tx_en;
                tx_en_2 <= tx_en_1;
                tx_en_3 <= tx_en_2;
            end
    wire tx_en_pos;					//Rising edge of tx_en
    assign tx_en_pos = tx_en_1 & tx_en_2 & (~tx_en_3);
            
    //Capture the rising edge of rx_en, reset the cnt count, and start generating the SPI clock		
    reg rx_en_1;
    reg rx_en_2;
    reg rx_en_3;
    always @(posedge clk or negedge rst_n)
        if (!rst_n)
            begin
                rx_en_1 <= 1'b0;
                rx_en_2 <= 1'b0;
                rx_en_3 <= 1'b0;
            end
        else 
            begin 
                rx_en_1 <= rx_en;
                rx_en_2 <= rx_en_1;
                rx_en_3 <= rx_en_2;
            end
    wire rx_en_pos;					//Rising edge of tx_en
    assign rx_en_pos = rx_en_1 & rx_en_2 & (~rx_en_3);		
            

    //Generate 100K SPI clock
    reg [8:0] cnt;
    always @(posedge clk or negedge rst_n)
        if(!rst_n)
            cnt <= 9'd0;
        else 
            begin
                if (rx_en == 1'b0 &&  tx_en == 1'b0)	
                    cnt <= 9'd0;
                if (cnt == 9'd499)
                    cnt <= 9'd0;
                if (rx_en == 1'b1 ||  tx_en == 1'b1)
                    cnt <= cnt + 1'b1;
            end
            
            
        // else if (tx_en_pos || rx_en_pos)
            // cnt <= 9'd0;
        // else if(cnt == 9'd499 && (tx_en | rx_en) == 1'b1)	//
            // cnt <= 9'd0;
        // else if(cnt < 9'd499 && (tx_en | rx_en) == 1'b1)		//The SPI clock signal is generated only when the transmission is enabled.
            // cnt <= cnt + 1'b1;
    reg sclk_r;
    always @(posedge clk or negedge rst_n)
        if (!rst_n)
            sclk_r <= 1'b1;
        else if (cnt == 9'd1)
            sclk_r <= 1'b0;
        else if (cnt == CLKSPI/2 - 1)
            sclk_r <= 1'b1;

    assign sclk = sclk_r;		
    //Count 10 clock cycles
    reg [3:0] cnt_delay10;
    always @(posedge clk or negedge rst_n)
        if (!rst_n) cnt_delay10 <= 4'd0;
        else if (rx_flag == 1'b1) cnt_delay10 <= cnt_delay10 + 1'b1;
        else if (rx_flag == 1'b0) cnt_delay10 <= 4'd0;
            
    //Send 1/2/4/6 bytes of data, use the variable length data transmission mode, but pass the number of bytes of the transmitted data
    reg [6:0] num;			//Send data count
    reg [4:0] num_rx;
    reg [6:0] num_tx;
    reg [13:0] num_bit;
    reg [2:0] num_byte;
    reg [3:0] byte_8;
    reg [7:0] data_8;
    always @(posedge clk or negedge rst_n)
        if (!rst_n)
            begin
                num <= 7'b0;
                num_rx <= 5'd0;
                num_tx <= 7'd0;
                rx_ready <= 1'd0;
                tx_ready <= 1'd0;
                mosi <= 1'b0;			
                sel <= 1'b1;
                num_bit <= 14'd0;
                num_byte <= 3'd0;
                rx_flag <= 1'b0;
                byte_8 <= 4'd0;
                data_8 <= 8'd0;
                rx_data_1 <= 8'd0;
                buff_8 <= 8'h00;
                rx_data_2 <= 16'h0000;
            end
        else 
            begin
                if (tx_en_pos)  //
                begin
                    tx_ready <= 1'b0;		//Once the rising edge is in the just-start state, the completion flag is set to 0
                    sel <= 1'b0;	
                end
                if (tx_en == 1'b1)
                begin
                case (tx_len)
                    3'd1: begin		//One byte of data	
                            if (cnt == 9'd1 && num < 7'd32)		//Add a sel == 1'b0 condition to ensure that sel is pulled low before sending data
                                begin
                                    num <= num + 1'b1;
                                    mosi <= tx_data[71-num];
                                    //tx_ready <= 1'b0;//change
                                end
                            if (cnt == 9'd251 && num == 7'd32) 
                                begin
                                    num <= 7'd0;
                                    tx_ready <= 1'b1;
                                    sel <= 1'b1;
                                end
                    end
                    
                    3'd2:	begin		//Two-byte data
                            if (cnt == 9'd1 && num < 7'd40)
                                begin
                                    num <= num + 1'b1;
                                    mosi <= tx_data[71-num];
                                    //tx_ready <= 1'b0;//change
                                end
                            if (cnt == 9'd251 && num == 7'd40) 
                                begin
                                    num <= 7'd0;
                                    tx_ready <= 1'b1;
                                    sel <= 1'b1;
                                end
                    end
                    
                    3'd4: begin		//Four Bytes Data
                            if (cnt == 9'd1 && num < 7'd56)
                                begin
                                    num <= num + 1'b1;
                                    mosi <= tx_data[71-num];
                                    //tx_ready <= 1'b0;//改动
                                end
                            if (cnt == 9'd251 && num == 7'd56) 
                                begin
                                    num <= 7'd0;
                                    tx_ready <= 1'b1;
                                    sel <= 1'b1;
                                end
                    end
                    
                    3'd6: begin		//Six bytes of data
                            if (cnt == 9'd1 && num < 7'd72)
                                begin
                                    num <= num + 1'b1;
                                    mosi <= tx_data[71-num];
                                    //tx_ready <= 1'b0;//Change
                                end
                            if (cnt == 9'd251 && num == 7'd72) 
                                begin
                                    num <= 7'd0;
                                    tx_ready <= 1'b1;
                                    sel <= 1'b1;
                                end
                        end		
                endcase
            end

                //Read 1/2 byte of data in the register
                if (rx_en_pos)
                begin 
                    rx_ready <= 1'b0;		//Once the rising edge is in the just-start state, the completion flag is set to 0
                    sel <= 1'b0;
                end
                if (rx_en == 1'b1)
                begin
                case (rx_len)
                    2'd0: begin	//Read all received data here					
                            if (cnt == 9'd1 && num < 7'd24)		
                                begin
                                    num <= num + 1'b1;
                                    mosi <= tx_data[71-num];
                                    //tx_ready <= 1'b0;
                                end
                            if (num_bit < rx_size * 8 && num == 7'd24 && cnt == 9'd1)
                                begin
                                    mosi <= 1'b0;
                                    num_bit <= num_bit + 1'b1;
                                    //num_byte <= num_byte + 1'b1;
                                end
                            if (num_bit < (rx_size * 8 + 1'b1)  && num == 7'd24 && num_bit > 14'd0 && cnt == 9'd250)
                                begin
                                    byte_8 <= byte_8 + 1'b1;
                                    data_8[7-byte_8] <= miso;
                                end
                            if (num_bit > 14'd0 && byte_8 == 4'd8 && rx_flag == 1'b0)
                                begin								
                                    rx_flag <= 1'b1;
                                    buff_8 <= data_8;		//Completes reception of one byte
                                    byte_8 <= 4'd0;
                                end
                            if (cnt_delay10 == 4'd10 && rx_flag == 1'b1)
                                begin
                                    rx_flag <= 1'b0;
                                end
                            if (cnt == 9'd255 && num_bit == rx_size * 8) 
                                begin
                                    num <= 7'd0;
                                    num_bit <=14'd0;
                                    rx_ready <= 1'b1;
                                    sel <= 1'b1;
                                    rx_flag <= 1'b0;
                                end						
                    end			
                    2'd1:	begin			
                            if (cnt == 9'd1 && num_tx < 7'd32)
                                begin
                                    num_tx <= num_tx + 1'b1;
                                    mosi <= tx_data[71-num_tx];
                                end
                            // if (cnt == 9'd251 && num_tx == 7'd32) 
                                // begin
                                    // num_tx <= 7'd0;
                                    
                                // end	
                            if (num_tx > 7'd24 && cnt == 9'd255 && num_rx < 5'd8)
                                begin
                                    num_rx <= num_rx + 1'b1;
                                    rx_data_1[7 - num_rx] <= miso;
                                end
                            if (num_rx == 5'd8)
                                begin
                                    num_tx <= 7'd0;
                                    num_rx <= 5'd0;
                                    rx_ready <= 1'b1;
                                    sel <= 1'b1;
                                end					
                    end
                    
                    2'd2:	begin
                            if (cnt == 9'd1 && num_tx < 7'd40)
                                begin
                                    num_tx <= num_tx + 1'b1;
                                    mosi <= tx_data[71-num_tx];
                                end
                            if (cnt == 9'd251 && num_tx == 7'd40) 
                                begin
                                    num_tx <= 7'd0;
                                    //rx_ready <= 1'b0;
                                end	
                            if (num_tx > 7'd24 && cnt == 9'd250 && num_rx < 5'd16)
                                begin
                                    num_rx <= num_rx + 1'b1;
                                    rx_data_2[15 - num_rx] <= miso;
                                end
                            if (num_rx == 5'd16)
                                begin
                                    num_rx <= 5'd0;
                                    rx_ready <= 1'b1;
                                    sel <= 1'b1;
                                end
                        end
                endcase
                end
                    
            end
        

endmodule
